# -*- coding: utf-8 -*-

import argparse
import logging
import os
import re
import signal
import string
import sys
import textwrap
from math import ceil

from colorama import Fore, Style
from requests import ConnectionError
from requests.packages import urllib3

from sbgrid import CLI_NAME, APP_NAME, ORG_NAME, __version__ as VERSION
from sbgrid.core import Installation, NotActivatedException, DamagedInstallationException

urllib3.disable_warnings()


def get_tty_size():
    """Get terminal size.

    Return:
        (int, int): height, width of terminal.
    """
    return map(int, os.popen('stty size', 'r').read().split())


def tabulate(packages, width):
    max_length = max(map(lambda p: len(p.slug), packages))
    number_of_columns = int(float(width + 1) / (max_length + 1))
    if number_of_columns == 0:
        # TODO !!! Test on bch?
        return list(packages[0].slug)
    chunk = int(ceil(float(len(packages)) / number_of_columns))
    columns = [packages[i:i + chunk] for i in range(0, len(packages), chunk)]
    lines = list()
    for i in range(0, chunk):
        line = list()
        for column in columns:
            if len(column) > i:
                if column[i].installed:
                    line.append(green(column[i].slug.ljust(max_length)))
                else:
                    line.append(column[i].slug.ljust(max_length))
        lines.append(' '.join(line))
    return lines


def green(s):
    return Fore.GREEN + Style.BRIGHT + s + Style.RESET_ALL


def magenta(s):
    return Fore.MAGENTA + Style.BRIGHT + s + Style.RESET_ALL


class Installer(object):

    def __init__(self):
        parser = argparse.ArgumentParser(prog=CLI_NAME, description=APP_NAME, usage='''
{0} <command> [<command_args>]

Available commands are:
    activate      activate new installation
    grids         list available grids
    info          print package metadata
    install       install single package
    install-all   install all packages from collection
    installed     list installed packages
    list          list available packages
    rebuild       rebuild installation
    remove        remove installed package
    resume        resume interrupted installations
    update        update package to latest release/version
    update-all    update all packages to latest release/version
    updates       list available updates for installed packages
    version       print package version string
'''.format(CLI_NAME), add_help=False)
        parser.add_argument('-h', '--help', help='show usage infromation and exit', action='store_true', default=False)
        parser.add_argument('--debug', help='enable verbose logging', action='store_true', default=False)
        parser.add_argument('command', help='subcommand to run')
        (args, command_args) = parser.parse_known_args()
        if args.debug:
            logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
        else:
            logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(message)s')
        for handler in logging.root.handlers:
            handler.addFilter(logging.Filter('root'))
        args.command = args.command.replace('-', '_')

        if args.command == 'version':
            logging.info(VERSION)
            sys.exit(0)
        if not hasattr(self, args.command):
            logging.info('Unrecognized command')
            parser.print_help()
            sys.exit(1)

        if args.help:
            command_args.append('--help')

        self.height, self.width = get_tty_size()
        self.installation = Installation()
        self.installation.debug = args.debug

        try:
            logging.debug('Loading configuration...')
            self.installation.load()
        except ConnectionError:
            logging.fatal('Need Internet connection to continue! Exiting...')
            sys.exit(1)
        except NotActivatedException:
            if args.command != 'activate':
                logging.error('Installation was not activated!')
                sys.exit(1)
        except DamagedInstallationException, e:
            logging.error('''{0} Installation is damaged! {1}'''.format(ORG_NAME, str(e)))
            sys.exit(1)

        os.setpgrp()
        try:
            getattr(self, args.command)(command_args)
        except ConnectionError:
            logging.fatal('Need Internet connection to continue! Exiting...')
            sys.exit(1)
        except KeyboardInterrupt:
            os.killpg(0, signal.SIGKILL)
        except Exception, e:
            logging.error(e.message, exc_info=args.debug)

    def _validate_grid_name(self, grid):
        for g in self.installation.grids:
            if grid.lower() == g.lower():
                return g
        raise Exception('Unknown grid name!')

    def _validate_collection_names(self, grid, collection):
        if collection and collection != 'all':
            name = ' '.join(collection).strip().lower()
            for c in self.installation.collections[grid]:
                if c.lower() == name.lower():
                    return ((grid, [c]),)
            raise Exception('Unknown collection name!')
        return ((grid, self.installation.collections[grid]), )

    def activate(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} activate'.format(CLI_NAME), description='Activate new installation.')
        parser.add_argument('site', help='site name')
        parser.add_argument('user', help='user name')
        parser.add_argument('key', help='activation key')
        args = parser.parse_args(command_args)
        if (self.installation.activate(args.site, args.user, args.key)):
            logging.info('Activation successfull!')
        else:
            logging.info('Activation failed!')

    def list(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} list'.format(CLI_NAME), description='List packages.')
        parser.add_argument('--grid', help='grid name', default=CLI_NAME)
        parser.add_argument('collection', nargs='*', help='collection name')
        args = parser.parse_args(command_args)
        logging.debug(args)
        for grid, collections in self._validate_collection_names(self._validate_grid_name(args.grid), args.collection):
            for collection in collections:
                logging.info(' -- {0} {1}'.format(magenta(collection), (self.width - len(collection) - 5) * '-'))
                for line in tabulate(self.installation.packages_by_collection(grid, collection), self.width):
                    logging.info(line)

    def _install_with_dependencies(self, package, version):
        for s, v in version.depends_on:
            pkg = self.installation.package(s)
            self._install_with_dependencies(pkg, pkg.version(v))
        self.installation.install(package, version)

    def install(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} install'.format(CLI_NAME), description='Install package.')
        parser.add_argument('slug', help='package name')
        parser.add_argument('version', nargs='?', help='package version', default=None)
        args = parser.parse_args(command_args)
        package = self.installation.package(args.slug)
        if args.version is None:
            logging.info('Installing default version')
            version = package.default_version
        else:
            version = package.version(args.version)
        self._install_with_dependencies(package, version)
        self.installation.setup()

    def install_all(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} install-all'.format(CLI_NAME),
                                         description='Install all packages in selected collection.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        parser.add_argument('--grid', help='grid name', default=CLI_NAME)
        parser.add_argument('collection', nargs='*', help='collection name', default='all')
        args = parser.parse_args(command_args)
        logging.debug(args)
        packages_to_install = list()
        for grid, collections in self._validate_collection_names(self._validate_grid_name(args.grid), args.collection):
            for collection in collections:
                for package in self.installation.packages_by_collection(grid, collection):
                    packages_to_install.append(package)
        if not args.yes:
            logging.info('Packages to be installed:')
            for package in packages_to_install:
                logging.info('%s (%s)', package, package.default_version)
            logging.info('Continue? [Y/n]?')
            if raw_input().strip().lower().startswith('n'):
                return
        for package in packages_to_install:
            self._install_with_dependencies(package, package.default_version)
        self.installation.setup()

    def remove(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} remove'.format(CLI_NAME), description='Remove package.')
        parser.add_argument('slug', help='package name')
        parser.add_argument('version', nargs='?', help='package version', default=None)
        args = parser.parse_args(command_args)
        package = self.installation.package(args.slug)
        if args.version is None:
            self.installation.remove_package(package)
        else:
            version = package.version(args.version)
            self.installation.remove(package, version)
        logging.info('Package removed!')
        self.installation.setup()

    def resume(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} resume'.format(CLI_NAME),
                                         description='Resume interrupted installations.')
        parser.parse_args(command_args)
        logging.info('Resuming interrupted installations...')
        for slug, version in self.installation.pending_packages():
            package = self.installation.package(slug)
            logging.info('Installing: {0} ({1})'.format(slug, version))
            self._install_with_dependencies(package, package.version(version))
        self.installation.setup()

    def installed(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} installed'.format(CLI_NAME), description='List installed packages.')
        parser.parse_args(command_args)
        if not self.installation.installed_packages():
            return
        logging.info('Installed packages:')
        for package in self.installation.installed_packages():
            logging.info('{0} ({1})'.format(package.slug, ', '.join([v.version for v in package.versions if v.installed])))  # noqa

    def info(self, command_args):
        def format_long_text(s):
            s = s.split('</p>')
            s = map(string.strip, s)
            s = map(lambda a: a.replace('</p>', '\n'), s)
            s = map(lambda a: a.replace('\r', ''), s)
            s = map(lambda a: re.sub(r'<[^>]*>', '', a), s)
            s = map(lambda a: textwrap.fill(a, width=75, initial_indent=5 * ' '), s)
            return '\n'.join(s)

        parser = argparse.ArgumentParser(prog='{0} info'.format(CLI_NAME), description='Print package metadata.')
        parser.add_argument('slug', help='package name')
        args = parser.parse_args(command_args)
        try:
            package = self.installation.package(args.slug)
            logging.info('Package: %s', package.title)
            logging.info('Name slug: %s', package.slug)
            logging.info('\nAvailable versions:')
            for version in package.versions:
                extra_info = []
                if version.installed:
                    extra_info.append('installed')
                if version.default:
                    extra_info.append('default')
                if version.obsolete:
                    extra_info.append('obsolete')
                extra_info = extra_info and '({0})'.format(', '.join(extra_info)) or ''
                dependencies = []
                for p, v in version.depends_on:
                    dependencies.append('{0} ({1})'.format(p, v))
                dependencies = dependencies and 'depends on: {0}'.format(', '.join(dependencies)) or ''
                logging.info('\t* %s', ' '.join([a for a in (version.version, extra_info, dependencies) if a]))
            logging.info('\nGrids(s): %s', ', '.join(package.grids))
            if package.collections:
                logging.info('Collection(s): %s', ', '.join(package.collections))
            if package.description:
                logging.info('\nDescription:')
                logging.info(format_long_text(package.description))
            if package.technical_notes:
                logging.info('\nTechnical notes:')
                logging.info(format_long_text(package.technical_notes))
            if package.links:
                logging.info('\nLinks:')
                for name, link in package.links:
                    logging.info('\t{0}: {1}'.format(name.replace('_', ' ').capitalize(), link))
        except Exception, e:
            logging.error(e.message)

    def updates(self, command_args=None):
        parser = argparse.ArgumentParser(prog='{0} updates'.format(CLI_NAME),
                                         description='List available updates for installed packages.')
        parser.parse_args(command_args)
        for package in self.installation.updated_packages():
            for version in package.updated_versions:
                if version.new_version:
                    logging.info('{0} {1} -> {2}'.format(package.slug, version.version, version.new_version.version))
                else:
                    logging.info('{0} {1} {2} -> {3}'.format(package.slug, version.version,
                                                             version.installed_timestamp, version.timestamp))

    def update(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} upgrade'.format(CLI_NAME),
                                         description='Update package to latest release/version.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        parser.add_argument('slug', help='package name')
        args = parser.parse_args(command_args)
        package = self.installation.package(args.slug)
        if not package.installed:
            logging.fatal('Package %s is not installed!', args.slug)
            sys.exit(1)
        if not package.update:
            logging.fatal('No updates for %s!', args.slug)
            sys.exit(1)
        for version in package.updated_versions:
            if not args.yes:
                if version.new_version:
                    logging.info('Package %s %s will be upgraded to version %s. ', package.slug,
                                 version.version, version.new_version.version)
                else:
                    logging.info('Package %s %s %s will be upgraded to new release %s. ', package.slug,
                                 version.version, version.installed_timestamp, version.timestamp)
                logging.info('Continue? [Y/n]?')
                if raw_input().strip().lower().startswith('n'):
                    continue
            self.installation.update(package, version)
        logging.info('Package upgraded successfully!')
        self.installation.setup()

    def update_all(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} upgrade-all'.format(CLI_NAME),
                                         description='Update all packages to latest release/version.')
        parser.add_argument('--yes', help='continue without confirmation', action='store_true')
        args = parser.parse_args(command_args)
        if not args.yes:
            logging.info('All packages will be upgraded to latest versions.')
            logging.info('Continue? [Y/n]?')
            if raw_input().strip().lower().startswith('n'):
                sys.exit(0)
        for package in self.installation.updated_packages():
            for version in package.updated_versions:
                self.installation.update(package, version)
        logging.info('All packages upgraded successfully!')
        self.installation.setup()

    def rebuild(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} rebuild'.format(CLI_NAME), description='Rebuild installation.')
        parser.parse_args(command_args)
        logging.warning('Rebuilding whole installation!')
        logging.warning('This might take a LONG time depending on number of missing/damaged packages!')
        self.installation.rebuild()
        self.installation.setup()
        logging.info('Rebuild completed successfuly')

    def grids(self, command_args):
        parser = argparse.ArgumentParser(prog='{0} grids'.format(CLI_NAME), description='List available grids.')
        parser.parse_args(command_args)
        for grid in self.installation.grids:
            logging.debug((grid.lower(), CLI_NAME))
            if grid.lower() == CLI_NAME:
                logging.info(green(grid))
            else:
                logging.info(grid)


def main():
    Installer()


if __name__ == '__main__':
    main()
